@@ -1,7 +1,7 @@
 /*
   Copyright (C) 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003,
-  2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012 Free Software
-  Foundation, Inc.
+  2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013 Free
+  Software Foundation, Inc.
 
   This file is part of GNU Inetutils.
 
@@ -58,6 +58,7 @@
 #include <time.h>
 #include <signal.h>
 #include <stdio.h>
+#include <unused-parameter.h>
 #include "talk.h"
 
 #define MSG_INTERVAL 4
@@ -66,7 +67,7 @@ char *current_state;
 int current_line = 0;
 
 static void
-disp_msg (int sig)
+disp_msg (int sig _GL_UNUSED_PARAMETER)
 {
   message (current_state);
 }
